/**
  OpenGL/GLUT/GLUI "playground" for 2D mesh-based physics.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-02-23 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */
#include "osl/mat4.h" /* for easy 4x4 matrix manipulation */
#include "GL/glui.h" /* user interface widgets: http://glui.sourceforge.net/ */

#include "soil/SOIL.h" /* Simple OpenGL Image Library, www.lonesock.net/soil.html (plus Dr. Lawlor's modifications) */
#include "soil/SOIL.c" /* just slap in implementation files here, for easier linking */
#include "soil/stb_image_aug.c"

#include "ogl/glsl.h"
#include "ogl/glsl.cpp"

#include "glui/glui.h"

#include "ogl/minicam.h" /* ortho_frame, camera motion, key_down & events */

#include <vector> /* std::vector container */
#include <sstream> /* for istringstream */
#include <map>

#ifndef M_PI /* why doesn't windows math.h have this? */
#define M_PI 3.14159265358979
#endif

/* These variables are changed directly by the user interface code below */

/* Enable for anisotropic filtering */
int texaniso=1;

/* Texture pixel format on graphics card */
int texformat=GL_RGBA8; 

/* Texture "wrap mode" in X and Y: */
int wrap_mode=GL_CLAMP_TO_EDGE; // REPEAT;

/* Texture magnification mode/minification mode */
int minfilter_mode=GL_LINEAR_MIPMAP_LINEAR; 
int magfilter_mode=GL_LINEAR; 




/** This struct represents one uniform or constant variable in a GLSL program */
class GLSL_variable {
public:
	std::string name; /* name in GLSL code */
	int location; /* glGetUniformLocation of uniform */
	float value;  /* current value */
};

/** This class extracts variable variables from a GLSL shader,
and allows C++ to query them. */
class GLSL_variable_finder {
public:
	GLSL_variable_finder(std::string programText,GLUI *glui);
	
	/* Return the location of this GLSL uniform (or -1 if none exists) */
	int getUniformLocation(std::string variableName) {
		GLSL_variable *u=variables[variableName];
		if (u) return u->location;
		else return -1;
	}
	/* Return the value of this GLSL variable (or 0.0 if none exists) */
	float getVariableValue(std::string variableName) {
		GLSL_variable *u=variables[variableName];
		if (u) return u->value;
		else return 0.0;
	}
	
	/* Set the GLSL value of all our uniform variables */
	void setUniforms(GLhandleARB prog) {
		vars_t::iterator it=variables.begin();
		while (it!=variables.end()) {
			GLSL_variable *u=(*it).second;
			if (u->location==-2) u->location=glGetUniformLocationARB(prog,u->name.c_str());
			if (u->location>=0)
				glUniform1fARB(u->location,u->value);
			++it;
		}
	}
private:
	typedef std::map<std::string,GLSL_variable *> vars_t;
	vars_t variables;
};

GLSL_variable_finder::GLSL_variable_finder(std::string programText,GLUI *glui)
{
	std::istringstream prog(programText);
	std::string line;
	int nspinners=0;
	while (getline(prog,line)) {
		std::istringstream linestream(line);
		std::string kind,type,name; 
		linestream>>kind>>type>>name;
		if (kind=="uniform" || kind=="const") {
			size_t loc;
			loc=name.find(";");
			if (loc!=std::string::npos) name=name.substr(0,loc);
			loc=name.find("=");
			if (loc!=std::string::npos) name=name.substr(0,loc);
		
			std::cout<<"Read "<<kind<<" variable '"<<name<<"', a "<<type<<"\n";
			float value=0.0;
			loc=line.find("="); /* start of assignment */
			if (loc!=std::string::npos) { /* has an assignment */
				std::string strvalue=line.substr(loc+1);
				std::istringstream streamvalue(strvalue);
				streamvalue>>value;
			}
			GLSL_variable *u=new GLSL_variable;
			u->name=name;
			u->location=-1; /* meaning unknown & unknowable */
			u->value=value;
			variables[name]=u; /* add to our map */
			
			if (kind=="uniform" && type=="float")
			{
				u->location=-2; /* meaning "please look up" */
				new GLUI_Spinner(glui,name.c_str(),&u->value);
				nspinners++;
				if (nspinners%3==0) new GLUI_Column(glui);
			}
		}
	}
}

std::string pixelShader;
GLSL_variable_finder *vf;

int reset=0; /* if true, start the simulation from initial conditions */
int stop=0; /* if true, stop running the simulation */
int limit_framerate=1; /* if true, clamp the framerate to 100Hz */

void make_ui(GLUI *glui,const char *shaderFile)
{
	new GLUI_Checkbox(glui,"Reset Simulation",&reset);
	new GLUI_Checkbox(glui,"Stop Simulation",&stop);
	new GLUI_Checkbox(glui,"Limit Framerate",&limit_framerate);
	new GLUI_Column(glui);
	
	pixelShader=readFileIntoString(shaderFile);
	vf=new GLSL_variable_finder(pixelShader,glui);
}



/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %s at %s\n",gluErrorString(err),where);
		exit(1);
	}
}

/* Make an OpenGL texture handle, of this size, with this pixel format */
GLuint make_texture(int w,int h,GLenum format=GL_RGBA8)
{
	GLuint tex=0;
	glGenTextures(1,&tex);
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,tex);
	char *mem=new char[w*h*4]; memset(mem,0x80,w*h*4);
	glTexImage2D(t,0,format, w,h,0, GL_RGBA,GL_UNSIGNED_BYTE,mem);
	delete[] mem;
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,GL_LINEAR); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,GL_LINEAR); /* no mipmaps */
	GLenum texWrap=GL_CLAMP_TO_EDGE; /* stop changing at edge */
	glTexParameteri(t, GL_TEXTURE_WRAP_S, texWrap); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, texWrap);
	glGenerateMipmapEXT(t); /* make mipmaps */
	glBindTexture(t,0); /* back to default texture */
	oglCheck("after make_texture");
	return tex;
}

/* Make a framebuffer object that renders pixels into this texture */
GLuint make_fbo(GLuint tex)
{
	GLuint fb=0;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
	
	/* Make sure the framebuffer's happy */
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
	{ fprintf(stderr,"Error %d in framebuffer creation!\n",status); exit(1); }

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after make_fbo");
	return fb;
}

// Call glVertex, and glTexCoords using this matrix
void vertex_with_texture(const osl::mat4 &m,const vec3 &world)
{
	vec4 w(world.x,world.y,world.z,1.0);
	glTexCoord4fv(m*w);
	glVertex4fv(w);
}

void display(void) /* Draw everything in our window */
{
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float display_dt=0.0;
	float big=300; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	
/* Do offscreen rendering work */

	/* This matrix projects the last frame's world coordinates into texture coords */
	static osl::mat4 toTexCoords(1.0);

	/* load an image file directly as a new OpenGL texture */
	GLenum texture_format=GL_RGBA_FLOAT16_ATI; // GL_RGBA8; // or try GL_RGBA_FLOAT16_ATI;
	static GLuint srcTex = 0;
	if (srcTex==0 || reset) {
		glDeleteTextures(1,&srcTex);
		srcTex=SOIL_load_OGL_texture(
			"image.jpg",
			SOIL_LOAD_AUTO,
			SOIL_CREATE_NEW_ID,
			SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y,
			texture_format
		);
	}
	int w=1024, h=1024;
	static GLuint destTex = make_texture(w,h,texture_format);
	static GLuint fb = make_fbo(destTex);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	
  if (!stop) {
/* Grab the average values from the highest mipmap level */
	int level=0;
	int size=w;
	while (size!=1) { level++; size/=2; }
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, srcTex, level);
	float averages[4];
	glReadPixels(0,0,1,1,GL_RGBA,GL_FLOAT,averages);
	
/* Render next step of simulation */
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, destTex, 0);
	glViewport(0,0,w,h);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

/* Set up the main camera projection matrix */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.1,1000.0 /* Z range */
	);
	
	/* Physics and display timesteps: */
	oglCameraLookAt(0.5);

	oglCheck("After matrix stuff");
	
	glDisable(GL_BLEND);
	
	/* Actual physics happens right on the graphics card */
	static GLhandleARB prog = makeProgramObject(
	"varying vec3 worldCoords;\n"
	"void main(void) /* vertex shader */ {\n"
	"	vec4 mv = gl_ModelViewProjectionMatrix * gl_Vertex; \n"
	"	worldCoords = vec3(gl_Vertex); \n"
	"	gl_Position = mv;\n"
	"}\n"
	,
	pixelShader.c_str()
	);
	glUseProgramObjectARB(prog);
	glUniform1iARB(glGetUniformLocationARB(prog,"srcTex"),0); /* texture unit 0 */
	glUniform4fvARB(glGetUniformLocationARB(prog,"averages"),1,averages);
	glUniformMatrix4fv(glGetUniformLocationARB(prog,"toTexCoords"),
			1,0,&toTexCoords[0][0]);
			
	/* toTexCoords takes 3D points and transforms them into texture coordinates.
	  We do it *after* setting the uniform above, because we're reading from
	  the *previous* frame, not the current frame.
	*/
	glGetFloatv(GL_PROJECTION_MATRIX,&toTexCoords[0][0]);
	
	/* Convert from -1,1 to 0,1: basically premultiply by
		glTranslatef(0.5,0.5,0.0);
		glScalef(0.5,0.5,1.0);

		 x   y   z    w   (in) 
	x	[0] [4] [8] [12]
	y	[1]	[5] [9] [13]
	z	[2] [6] [10][14]
	w	[3] [7] [11][15]
	*/
	osl::mat4 scaleTrans(1.0);
	for (int i=0;i<2;i++) {
		scaleTrans[i][i]*=0.5;
		scaleTrans[3][i]+=0.5;
	}
	toTexCoords=scaleTrans*toTexCoords; 
	

	/* Figure out which directions to pull data from */
	int dirs=(int)vf->getVariableValue("dirs");
	if (dirs==0) dirs=4; /* don't know how many directions to use */
	static float *srcDirs=new float[2*dirs];
	for (int d=0;d<dirs;d++) {
		float r=1.0; /* filter size, in pixels */
		float angle=d*(2.0*M_PI/dirs); /* filter orientation, radians */
		float dx=r*cos(angle), dy=r*sin(angle);
		if (dirs==8) /* round to integer coordinates, not float */
			{dx=ceil(dx-0.5); dy=ceil(dy-0.5);}
		srcDirs[2*d+0]=dx*(1.0/w); /* to pixel coordinates */
		srcDirs[2*d+1]=dy*(1.0/w);
	}
	glUniform2fvARB(glGetUniformLocationARB(prog,"srcDirs"),dirs,srcDirs);
	oglCheck("After uniforms");
	
	vf->setUniforms(prog);
	
	/* run this program over the whole rectangle of the texture */
	glColor4f(1,1,1,1); 
	glBegin(GL_QUADS);
	glVertex3f(-big,-big,floor);
	glVertex3f(+big,-big,floor);
	glVertex3f(+big,+big,floor);
	glVertex3f(-big,+big,floor);
	glEnd();
	oglCheck("After draw");
	
	glUseProgramObjectARB(0);
	
	if (fmod(start_t,2.0)<1.0) { // key_down[(int)' ']) { /* interact with the simulation */
		glPushMatrix();
		float x=mouse_x/(float)glutGet(GLUT_WINDOW_WIDTH);
		float y=1.0-mouse_y/(float)glutGet(GLUT_WINDOW_HEIGHT);
		glTranslatef(x,y,0);
		glDisable(GL_TEXTURE_2D);
		glColor4f(1.0,1.0,1.0,1.0);
		glutSolidTeapot(0.1);
		glPopMatrix();
	}
	

	if (limit_framerate) {
	#ifdef _WIN32
		Sleep(10);
	#else
		usleep(10*1000); 
	#endif	
	}
	glDisable(GL_TEXTURE_2D);
	
	std::swap(srcTex,destTex); /* ping-pong the buffers */
	
	oglCheck("After offscreen work");
  } /* end if (!stop) */
  
/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
// Draw in our ground plane texture
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,srcTex);
	glEnable(t);
	glGenerateMipmapEXT(t); /* make mipmaps before rendering */
	
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,magfilter_mode); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,minfilter_mode);
	glTexParameteri(t, GL_TEXTURE_WRAP_S, wrap_mode); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, wrap_mode);
	glTexParameteri(t, GL_TEXTURE_MAX_ANISOTROPY_EXT,texaniso?16:1);
	
	glDisable(GL_BLEND);
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1); 
	vertex_with_texture(toTexCoords,vec3(-big,-big,floor));
	vertex_with_texture(toTexCoords,vec3(+big,-big,floor));
	vertex_with_texture(toTexCoords,vec3(+big,+big,floor));
	vertex_with_texture(toTexCoords,vec3(-big,+big,floor));
	glEnd();
	glDisable(GL_TEXTURE_2D);
	
	oglCameraAxes();
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed+0.001;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	
	const char *shaderFile="shader.txt";
	if (argc>1) shaderFile=argv[1]; /* first argument: shader to run */
	
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	int win=glutCreateWindow("Physics From 'shader.txt'"); /* make a window */
	glewInit(); /* only AFTER create window! */
	
	// Make some popups & sliders (this is the only GLUI code you'll need!)
        GLUI *glui = GLUI_Master.create_glui_subwindow(win,GLUI_SUBWINDOW_TOP);
        make_ui(glui,shaderFile);
	
	glutDisplayFunc(display); /* here's how to draw the window */
	oglCameraInit();
	//camera=vec3(0.5,0.5,0.80);
	//camera_orient.y=vec3(0,1,0);camera_orient.z=vec3(0,0,1);
	camera=vec3(0.5,-1,0.80);
	camera_orient.y=vec3(0,0,1);camera_orient.z=vec3(0,-1,0.3);
	glutMainLoop(); /* start drawing */
}
